/*	$NetBSD$	*/

/*-
 * Copyright (c) 2001 Minoura Makoto.
 * Copyright (c) 1994 Charles M. Hannum.
 * Copyright (c) 1992, 1993 Erik Forsberg.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL I BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Synaptics TouchPad driver.
 * Intended to provide common functions among PS/2, serial and ADB.
 * Based on the specification available from Synaptics.
 *   (http://www.synaptics.com/support/dev_support.cfm)
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/ioctl.h>

#include <machine/bus.h>

#include <dev/ic/stpreg.h>
#include <dev/ic/stpvar.h>

#include <dev/wscons/wsconsio.h>
#include <dev/wscons/wsmousevar.h>

static char UNK[] = "Unknown sensor type";
static char *stp_idstring[] = {
	UNK,
	"TM41xx134 Standard TouchPad",
	"TM41xx156 Mini module",
	"TM41xx180 Super module",
	UNK, UNK, UNK, UNK,
	"TM41xx220 Ultra-thin module",
	"TM41xx230 Wide pad module",
	UNK,
	"TM41xx240 Stamp pad module",
	"TM41xx140 SubMini module",
	UNK, UNK,
	"TM41xx301 Advanced Technology Pad",
	"TM41xx221 Ultra-thin module, connector reserved"
};
#define STP_NID		(sizeof(stp_idstring) / sizeof(stp_idstring[0]))

int stp_init (arg)
	void *arg;
{
	struct stp_softc *sc = arg;
	int r;
	stp_identifier *ident;
	struct wsmousedev_attach_args a;

	ident = &sc->sc_ident;
	r = (sc->sc_identify) (arg, ident);
	if (r) {
		printf("\nstp_init: identification failed\n");
		return EIO;
	}
	printf(": Synaptics TouchPad rev. %d.%d\n",
	       ident->si_info_major, ident->si_info_minor);

	r = ((ident->si_model_id & STP_MODELID_INFO_SENSOR_MASK)
	     >> STP_MODELID_INFO_SENSOR_SHIFT);
	printf("%s: %s", sc->sc_dev.dv_xname,
	       (r < STP_NID)?stp_idstring[r]:UNK);
	r = ((ident->si_model_id & STP_MODELID_INFO_GEOMETRY_MASK)
	     >> STP_MODELID_INFO_GEOMETRY_SHIFT);
	if (r == 1)
		printf(", rectangular");
	else if (r == 2)
		printf(", oval");
	if (ident->si_model_id & STP_MODELID_INFO_NEW_ABS)
		printf(", new packet");
	if (ident->si_model_id & STP_MODELID_CAP_PEN)
		printf(", pen capable");
	printf(".\n%s: setting absolute mode.\n", sc->sc_dev.dv_xname);
	r = (sc->sc_setmode) (arg, STP_MODE_ABSOLUTE);
	if (r) {
		printf("stp_init: setmode failed\n");
		return EIO;
	}
	sc->sc_seqno = 0;
	sc->sc_buttons = 0;
	sc->sc_status = STP_IDLE;

	a.accessops = sc->sc_wsaccess;
	a.accesscookie = arg;

	/*
	 * Attach the wsmouse, saving a handle to it.
	 */
	sc->sc_wsmousedev = config_found(arg, &a, wsmousedevprint);

	return 0;
}

void stp_input (arg, p)
	void *arg;
	struct stp_abspacket *p;
{
	struct stp_softc *sc = arg;
	int buttons = 0;
	int foundinput = 0;
	int x, y, dx = 0, dy = 0, i;

	memcpy(&sc->sc_history[1], &sc->sc_history[0],
	       sizeof (sc->sc_history) - sizeof (sc->sc_history[0]));
	sc->sc_history[0] = *p;

	/* Process buttons */
	if (sc->sc_history[7].sa_lbutton)
		buttons |= 1;
	if (sc->sc_history[7].sa_rbutton)
		buttons |= 4;
	if (sc->sc_buttons ^ buttons) {
		sc->sc_buttons = buttons;
		foundinput++;
	}

	/* Process movement */
	switch (sc->sc_status) {
	case STP_IDLE:
		i = x = y = 0;
		if (sc->sc_history[7].sa_z >= 80) {
			x += sc->sc_history[7].sa_x;
			y += sc->sc_history[7].sa_y;
			i++;
		}
		if (sc->sc_history[8].sa_z >= 80) {
			x += sc->sc_history[8].sa_x;
			y += sc->sc_history[8].sa_y;
			i++;
		}
		if (sc->sc_history[9].sa_z >= 80) {
			x += sc->sc_history[9].sa_x;
			y += sc->sc_history[9].sa_y;
			i++;
		}
		x /= i;
		if (x)
			dx = sc->sc_x - x;
		y /= i;
		if (y)
			dy = sc->sc_y - y;
		if (dx || dy)
			found++;
	}

	if (found)
		wsmouse_input(sc->sc_wsmousedev,
			      sc->sc_buttons, dx, dy, 0,
			      WSMOUSE_INPUT_DELTA);
}
